/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 29/08/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_opts_h__
#define __mc_engine_opts_h__ 1


/*   std includes   */
#include <map>
#include <vector>
#include <climits>
#include <iostream>

/*   project headers   */
#include <bits/duct-tags.h>



namespace duct {
 
  enum class evolution_type : int {
    typeI = 1,
    typeII = 2
  };
 
  /*   Simple structure to provide input parameters for the shower engine.   */
  struct shower_engine_opts
  {
    /*   for test purpuses we might want to use kind of insane options    */
    bool is_insane = false;
    
    /*   cutoff kT scale   */
    double kT_cutoff = 1.0_GeV;

    /*   max number of the basis state to consider in every steps   */
    unsigned int max_number_of_color_basis_states = 2u;

    
    
    /*   parameters to control the soft corrections    */
    struct soft_opts
    {
      soft_opts(unsigned ord = 0u, unsigned re = 0u, unsigned im = 0u)
      {
        order = std::min(ord, re+im);
        real = std::min(order, re);
        phase = std::min(order, im);
      }
      
      /*   control parameters    */
      unsigned order, real, phase;
    };
    
    
    /*   parameters to control the color approximations    */
    struct color_opts
    {
      color_opts(unsigned step = UINT_MAX, unsigned suppr = 0u, const soft_opts& soft = {},
                 evolution_type type = evolution_type::typeI, unsigned mxst = 2u)
      {
        max_number_of_steps = std::max(step, soft.real);
        max_color_suppression = suppr;
        soft_corrections = soft;
        type_of_evolution = type;
        max_number_of_basis_states = mxst;
      }
      
      
      /*   max number of steps with this approximations    */
      unsigned int max_number_of_steps;

      
      /*   max allowed color suppression   */
      unsigned int max_color_suppression;
      
      
      /*   We need some variables to control the approximation. We want to keep tracking the
       *   number of the insertions of Delta H and Delta V operators and the number of
       *   the insertions of the phase term.
       */
      soft_opts soft_corrections;

      /*    Type of the evolution   */
      evolution_type type_of_evolution;
      
      /*   max number of the basis state to consider in every steps   */
      unsigned int max_number_of_basis_states;
    };
    
  
    /*   color option for consecutive evolution intervals   */
    std::map<double, std::vector<color_opts>> color;

  
    struct threshold_opts
    {
      /*   enamble threshold contributions   */
      bool enabled = true;
      
      /*  controling Delta V expansion in the threshold operator    */
      unsigned int order_of_soft_corrections = 0u;
    } threshold;
  
    
    /*   clean up the shower options   */
    shower_engine_opts& sanitize();
  };
  
  
  /*   fancy print   */
  std::ostream& operator<<(std::ostream&, const shower_engine_opts&);

}

#endif
