/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/28/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_hard_params_h__
#define __mc_hard_params_h__ 1



/*   std includes   */
#include <experimental/vector>

/*   project headers   */
#include<bits/duct-model.h>
#include <cxx/memory-resource.h>



namespace duct {
 
  /*   Represents a set of hard parameters.
   *   Here we have the simple variables for initial and final state.
   */
  template<class> struct hard_params_pod;
  
  template<>
  struct hard_params_pod<ini>
  {
    /*   model description   */
    model_ref mdl;
 
    /*   PDF auxillary data   */
    const shower_aux *aux;
    
    /*   index of the emitter parton   */
    int emitter;
    
    /*   actual hard parameters   */
    int fl;
    double eta, Q2, mT, vnull;
    
    /*   veto parameters   */
    double mTmax, mYmax;
  };
  
  
  template<>
  struct hard_params_pod<fin>
  {
    /*   model description     */
    model_ref mdl;

    /*   index of the emitter parton   */
    int emitter;
    
    /*   actual hard parameters   */
    int fl;
    bool is_decay;
    double mT, vnull;
    double pQ, rl, ymax, al, bl;
    
    /*   veto parameters   */
    double mTmax, mYmax;
  };
  
  
  /*   Represents a set of hard parameters.
   *   We have separate classes for intial and final state splittings.
   */
  template<class>
  struct hard_params;
  
  
  
  template<>
  struct hard_params<ini> : hard_params_pod<ini>
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::memory_resource*;
    
    
    /*   default constructors    */
    hard_params() { spectators.reserve(30);}
    explicit hard_params(allocator_type a) : spectators{a} { spectators.reserve(30);}
    
    
    /*   copy   */
    hard_params(const hard_params&) = default;
    hard_params& operator=(const hard_params&) = default;
    
    hard_params(const hard_params& x, allocator_type a)
    : hard_params_pod<ini>{x}
    , spectators{x.spectators, a}
    {}
    
    /*    move  */
    hard_params(hard_params&&) = default;
    hard_params& operator=(hard_params&&) = default;
    
    hard_params(hard_params&& x, allocator_type a)
    : hard_params_pod<ini>{std::move(x)}
    , spectators{std::move(x.spectators), a}
    {}
    
    
    /*   spectator variables   */
    struct spectator_vars
    {
      int spectator;
      double zk, xi, xT, cfac;
    };
    
    /*   considering all the possible spectators    */
    pmr::vector<spectator_vars> spectators;
  };
  
  

  template<>
  struct hard_params<fin> : hard_params_pod<fin>
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::memory_resource*;
    
    
    /*   default constructors    */
    hard_params() { spectators.reserve(30);}
    explicit hard_params(allocator_type a) : spectators{a} { spectators.reserve(30);}
    
    
    /*   copy   */
    hard_params(const hard_params&) = default;
    hard_params& operator=(const hard_params&) = default;
    
    hard_params(const hard_params& x, allocator_type a)
    : hard_params_pod<fin>{x}
    , spectators{x.spectators, a}
    {}
    
    /*    move  */
    hard_params(hard_params&&) = default;
    hard_params& operator=(hard_params&&) = default;
    
    hard_params(hard_params&& x, allocator_type a)
    : hard_params_pod<fin>{std::move(x)}
    , spectators{std::move(x.spectators), a}
    {}
    
    
    /*   spectator variables   */
    struct spectator_vars
    {
      int spectator;
      double xi, xT, cfac;
    };
    
    /*   considering all the possible spectators    */
    pmr::vector<spectator_vars> spectators;
  };
  

  
  
  /*   These classes represents the sets of hard parameters   */
  template<ProcessTag _Proc>
  struct parameters
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::memory_resource*;
    
    
    /*   default constructors    */
    parameters() = default;
    explicit parameters(allocator_type a) : fin{a} {}
    
    
    /*   copy   */
    parameters(const parameters&) = default;
    parameters& operator=(const parameters&) = default;
    
    parameters(const parameters& x, allocator_type a)
    : fin{x.fin, a} {}
    
    
    /*    move  */
    parameters(parameters&&) = default;
    parameters& operator=(parameters&&) = default;
    
    parameters(parameters&& x, allocator_type a)
    : fin{std::move(x.fin), a} {}
    
    /*   parameters of the final state emitters   */
    pmr::vector<hard_params<duct::fin>> fin;
  };
  
  
  template<ProcessTag _Proc>
  requires HadronCollision<_Proc>
  struct parameters<_Proc>
  {
    /*    it has to be an allocator aware container     */
    using allocator_type = pmr::memory_resource*;
    
    
    /*   default constructors    */
    parameters() = default;
    explicit parameters(allocator_type a) : ini{a}, fin{a} {}
    
    
    /*   copy   */
    parameters(const parameters&) = default;
    parameters& operator=(const parameters&) = default;
    
    parameters(const parameters& x, allocator_type a)
    : ini{x.ini, a}, fin{x.fin, a} {}
    
    
    /*    move  */
    parameters(parameters&&) = default;
    parameters& operator=(parameters&&) = default;
    
    parameters(parameters&& x, allocator_type a)
    : ini{std::move(x.ini), a}, fin{std::move(x.fin), a} {}
    
    
    /*   parameters of the initial and final state emitters   */
    pmr::vector<hard_params<duct::ini>> ini;
    pmr::vector<hard_params<duct::fin>> fin;
  };
  
  
}    //    namespace duct

#endif

