/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/28/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_vz_mapping_h__
#define __mc_vz_mapping_h__ 1


/*   std includes   */
#include <algorithm>

/*   project headers   */
#include <bits/mc-ordering.h>
#include <bits/mc-hard-params.h>
#include <cxx/math-support.h>


namespace duct {

  namespace __details {
    
 
    /*   z mapping for the intial state
     *   ==> INITIAL state
     */
    class z_mapping_ini
    {
    public:
      /*   refresing the v dependence    */
      template<class _Ordering>
      void bind(const hard_params<ini> *pars, double v) {
        _M_pars = pars; _M_v = v;
        _M_zmax = ordering_traits<ini, _Ordering>::main::zmax(_M_pars, _M_v);
      }

      /*   mapping chi ==> z   */
      std::pair<double, double> w2z(splitting_type idx, double w) const {
        return _M_pars->aux->inverse(w, idx, _M_zmax, -_M_pars->fl, _M_pars->eta, _M_pars->Q2*_M_v);
      }
      
      /*   returns the zmin/zmax values  */
      double zmin() const { return _M_pars->eta;}
      double zmax() const { return _M_zmax;}
      
    private:
      /*   data members  */
      const hard_params<ini> *_M_pars;
      
      /*  cached variables   */
      double _M_v, _M_zmax;
    };
 
    
    /*   This class represents a simple spline based mapping for the ordering variable.
     *   It helps to implement the class splitting_functions_approx.
     */
    template<class _Type>
    class v_mapping;
    
    
    /*   v mapping for the intial state
     *   ==> INITIAL state
     */
    template<>
    class v_mapping<ini>
    {
      static constexpr unsigned int _N_grid = 20U;
      
    public:
      /*   generating tau  */
      std::pair<bool,double> generate_tau(double, double, double) const;
      
      /*    allowed kinematical range in variable v    */
      double vmin() const { return _M_vmin;}
      double vmax() const { return _M_vmax;}
      
    protected:
      /*    constructors are protected (it has to be a base class)   */
      v_mapping() = default;
      v_mapping(v_mapping&&) = default;
      v_mapping(const v_mapping&) = default;
      
      /*   binding to the current hard configuration    */
      template<OrderingTag _Ordering, class _Integrand>
      bool bind(_Ordering, const hard_params<ini> *pars, double, double, _Integrand&&);
      
      /*   hard parameters   */
      const hard_params<ini> *_M_pars;

      /*   heavy flavor threshold in tau  */
      double _M_tauT;
      
    private:
      /*   using spline to do the importance sampling  */
      spline<spline_type::linear> _M_spline;
      
      /*   limits   */
      double _M_tmin, _M_tmax;
      
      /*    allowed kinematical range in variable v    */
      double _M_vmin, _M_vmax;
    };
    
    
    template<OrderingTag _Ordering, class _Integrand>
    bool v_mapping<ini>::bind(_Ordering, const hard_params<ini> *pars, double tauL, double tauH, _Integrand&& integrand)
    {
      /*   current hard parameters   */
      _M_pars = pars;
      
      /*   aliases   */
      const auto& as = _M_pars->mdl.alpha_qcd();
      const shower_aux *aux = _M_pars->aux;
      const double &Q2 = _M_pars->Q2;
      
      /*  Available phase space in v variable.   */
      double vl = 1000.0, vh = 0.0;
      
      ordering_traits<ini, _Ordering>::main::vlimits(vl, vh, _M_pars);
      if(vl >= vh) return false;
      
      _M_vmin = vl;
      _M_vmax = vh;
      
      double q2T = aux->thrsT(_M_pars->fl, _M_pars->eta);
      double q2max = 0.99*Q2*vh, q2min = std::max({aux->q2min(), q2T, 1.01*Q2*vl});
      
      if(q2min >= q2max) return false;
      
      /*   kinematical limits in variable tau    */
      double kinL = as.tauk(q2max), kinH = as.tauk(q2min);
      
      /*   Threshold and limits in variable tau   */
      _M_tauT = as.tauk(std::max(q2T, aux->q2min()*0.99));
      _M_tmin = std::max(kinL, tauL);
      _M_tmax = std::min(kinH, tauH);
      
      if(_M_tmin >= _M_tmax) return false;
      
      
      /*   setting up a generic lambda to make grid in the splines    */
      double tauM = _M_tmin-1e-6;
      
      if constexpr(std::is_same_v<_Ordering, ordering::lambda>) {
        double mX = aux->thrsT(_M_pars->fl)/Q2;
        tauM = as.tauk(Q2*(0.25*sqr(std::sqrt(mX) + std::sqrt(mX + 4.0))-1.0));
      }
      
      else if constexpr(std::is_same_v<_Ordering, ordering::angular>) {
        double mX = aux->thrsT(_M_pars->fl)/Q2;
        tauM = as.tauk(0.25*sqr(std::sqrt(mX) + std::sqrt(mX + 4.0))*Q2);
      }
      
      /*   setting up the spline   */
      std::size_t nP = std::round(_N_grid*(_M_tmax-_M_tmin)/(kinH-kinL));
      
      if(_M_tmin < tauM && tauM < _M_tmax)
        _M_spline.makegrid(nP, {_M_tmin, tauM, _M_tmax});
      else _M_spline.makegrid(nP, _M_tmin, _M_tmax);
      
 
      /*   initializing the spline   */
      _M_spline.assign(std::forward<_Integrand>(integrand));
      
      
      return (_M_spline.assign_integral(logweight{_M_tauT}) > 1.0e-8 ? true : false);
    }

    
 
    
    /*   z mapping for the final state
     *   ==> FINAL state
     */
    template<OrderingTag>
    class z_mapping_fin;
    
    /*     Specialization for lambda ordering      */
    template<>
    class z_mapping_fin<ordering::lambda>
    {
    public:
      /*   refresing the v dependence    */
      void bind(const hard_params<fin> *, double);
      
      /*   mapping chi ==> z   */
      std::pair<double, double> w2z(splitting_type, double) const;
      
      /*   returns the zmin/zmax values  */
      double zmin() const { return 1.0-zh;}
      double zmax() const { return zh;}
      
    private:
      /*  cached variables   */
      double zh, ox0, asb, Lm, J;
    };
    
    
    /*     Specialization for lambda ordering      */
    template<>
    class z_mapping_fin<ordering::kT>
    {
    public:
      /*   refresing the v dependence    */
      void bind(const hard_params<fin> *, double);
      
      /*   mapping chi ==> z   */
      std::pair<double, double> w2z(splitting_type, double) const;
      
      /*   returns the zmin/zmax values  */
      double zmin() const { return 1.0-zh;}
      double zmax() const { return zh;}
      
    private:
      /*  cached variables   */
      double zh, ox0, Lm, J;
    };
    
    
    template<>
    class z_mapping_fin<ordering::angular>
    {
    public:
      /*   refresing the v dependence    */
      void bindL(const hard_params<fin> *, double);
      void bindH(const hard_params<fin> *, double);
      
      /*   mapping chi ==> z   */
      std::pair<double, double> w2z(splitting_type, double) const;
      
      /*   returns the zmin/zmax values  */
      double zmin() const { return zl;}
      double zmax() const { return zh;}
      
    private:
      /*  cached variables   */
      double zl, zh, asb, Lm, J;
    };


    
    /*   v mapping for the final state
     *   ==> FINAL state
     */
    template<>
    class v_mapping<fin>
    {
      /*   size of the interpolation grid    */
      static constexpr unsigned int _N_grid = 20U;
      
    public:
      /*   generating tau  */
      std::pair<bool,double> generate_tau(double, double, double) const;
         
      /*    allowed kinematical range in variable v    */
      double vmin() const { return _M_vmin;}
      double vmax() const { return _M_vmax;}
      
    protected:
      /*    constructors are protected (it has to be a base class)   */
      v_mapping() = default;
      v_mapping(v_mapping&&) = default;
      v_mapping(const v_mapping&) = default;
      
      
      /*   binding to the current hard configuration    */
      template<OrderingTag _Ordering, class _Integrand>
      bool bind(_Ordering, const hard_params<fin> *, double, double, _Integrand&&);
      
      /*   hard parameters   */
      const hard_params<fin> *_M_pars;
      
    private:
      /*    interpolation range and the whole integral  */
      double _M_tmin, _M_tmax;
      
      /*    allowed kinematical region in variable v   */
      double _M_vmin, _M_vmax;
      
      /*   spline for the tau grid   */
      spline<spline_type::linear> _M_spline;
    };
    
    
    template<OrderingTag _Ordering, class _Integrand>
    bool v_mapping<fin>::bind(_Ordering, const hard_params<fin> *pars, double tauL, double tauH, _Integrand&& integrand)
    {
      /*   current hard parameters   */
      _M_pars = pars;
      
      /*   aliases   */
      const auto& as = _M_pars->mdl.alpha_qcd();
      const double &pQ = pars->pQ;
      
      
      /*  Available phase space in v variable.   */
      double vl, vh;
      
      ordering_traits<fin, _Ordering>::main::vlimits(vl, vh, _M_pars);
      if(vl >= vh) return false;
      
      _M_vmin = vl;
      _M_vmax = vh;
      
      /*   Kinematical limits in tau variable   */
      double kinL = as.tauk(pQ*vh)*1.005, kinH = as.tauk(pQ*vl)*0.995;
     
      /*   Setting up the interpolation spline */
      _M_tmin = std::max(kinL, tauL);
      _M_tmax = std::min(kinH, tauH);
      
      if(_M_tmin >= _M_tmax) return false;
      
      /*   setting up the spline   */
      std::size_t nP = std::round(_N_grid*(_M_tmax-_M_tmin)/(kinH-kinL));
     
      _M_spline.makegrid(nP, _M_tmin, _M_tmax);
      _M_spline.assign(std::forward<_Integrand>(integrand));
      
  
      return (_M_spline.assign_integral() > 1.0e-8 ? true : false);
    }
  
  }  /*   namespace __details    */
  
  
  
  
}   /*  namespace duct   */

#endif
