/*  Copyright (C) 2012-2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __qcd_color_overlap_h__
#define __qcd_color_overlap_h__ 1


/*   std includes   */
#include <list>
#include <cmath>
#include <vector>


/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/mc-color-state.h>



namespace duct {
  
  /*   forward declaration of color overlap classes   */
  class color_overlap_su3;

  
  namespace __details {
    
    class color_graph_su3
    {
    public:
      /*    we use the __partile_state color link type   */
      using colorlink = parton_cstate::colorlink;
      
      /*    class for representing a color state      */
      struct parton
      {
        /*  Multiplicity of the parton insertion:
         *    => multi == 0 inactive removed from the list
         *    => multi > 0 the multiplicity of the parton
         *       (we can have many parallel gluon)
         */
        unsigned int multi;
        
        /* These members for gluons. We use multi-gluons and these
         * are the coefficients of the corresponding Fierz identity.
         */
        double A, B;
        
        /*    color connections    */
        colorlink color, colorP;
      };
      
      /*    default constructor    */
      color_graph_su3() : _M_number_of_in{0u}, _M_active{0u}, _M_cf_pow{0}, _M_nc_pow{0}, _M_wgt{1.0} {}
      
      /*   non standard constructor, creating from event<...>    */
      template<ProcessTag _Proc>
      color_graph_su3(const color_state<_Proc>& c)
      : _M_number_of_in{c.number_of_in()}
      {
        this->assign(c.begin(), c.begin()+_M_number_of_in, c.end());
      }
      
      /*   assignement operator from event   */
      template<ProcessTag _Proc>
      color_graph_su3& operator=(const color_state<_Proc>& c)
      {
        _M_number_of_in = c.number_of_in();
        this->assign(c.begin(), c.begin()+_M_number_of_in, c.end());
        return *this;
      }
      
      /*   returns true if it is done and the overlap is calculated   */
      bool is_finished() const { return !_M_active;}
      
      /*   this applys all the simple rules   */
      void reduce();
      
    private:
      /*   assigment functions   */
      void assign(const parton_cstate *, const parton_cstate *, const parton_cstate *);
      
      /*   rules to simplify the graps   */
      void ruleG();
      void ruleQ();
      void ruleA();
      void ruleT();
      void ruleS();
      void ruleF(color_graph_su3&);
      
      /*   some helper   */
      parton * zero() { return _M_graph.data() + _M_number_of_in - 1;}
      parton * start() { return _M_graph.data();}
      parton * end() { return _M_graph.data() + _M_graph.size();}
      
    private:
      /*    storage for the partons   */
      std::vector<parton> _M_graph;
      
      /*   number of incoming colored objects and active partons   */
      unsigned int _M_number_of_in, _M_active;
      
      /*   powers of the Cf and Nc factors   */
      int _M_cf_pow, _M_nc_pow;
      
      /*   weight of the graph   */
      double _M_wgt;
      
      /*   friend declaration   */
      friend class duct::color_overlap_su3;
    };
    
  }   //   namespace __details
  
  
  
  /*    callable object class for calculating SU(3) color overlap   */
  class color_overlap_su3
  {
  public:
    /*    calculate the weight of a color configuration    */
    template<ProcessTag _Proc>
    double operator()(const color_state<_Proc>& p) {
      return this->calculate(__details::color_graph_su3(p));
    }

    /*   calculate the overlap of a given graph   */
    double calculate(__details::color_graph_su3&&);

  private:
    /*     we use list to store the graphs    */
    std::list<__details::color_graph_su3>::iterator _M_begin, _M_end;
    std::list<__details::color_graph_su3> _M_graphs;
  };

  
  
  /*    callable object class for calculatig the color overlap     */
  class color_overlap
  {
  public:
    /*   calculate the U(3) overlap  */
    template<ProcessTag _Proc>
    double operator()(const color_state<_Proc>& c)
    {
      if(c.is_color_u3)
        return std::pow(qcd::Nc, -color_suppression_index_u3(c));
      
      return _M_overlap_su3(c);
    }
    
    /*   converting the color state to U(3)   */
    template<ProcessTag _Proc>
    void convert_to_u3(color_state<_Proc>& c)
    {
      if(c.is_color_u3 == false)
      {
        c.weight *= _M_overlap_su3(c);
        c.weight *= std::pow(qcd::Nc, color_suppression_index_u3(c));
        c.is_color_u3 = true;
      }
    }
    
    template<ProcessTag _Proc>
    void convert_to_u3(unsigned int max_color, color_state<_Proc>& c)
    {
      if(c.is_color_u3 == false && c.color_suppression_index >= max_color)
      {
        c.weight *= _M_overlap_su3(c);
        c.weight *= std::pow(qcd::Nc, color_suppression_index_u3(c));
        c.is_color_u3 = true;
      }
    }
    
   
  private:
    /*   underlying engines to calculate the overlaps    */
    color_overlap_su3 _M_overlap_su3;
  };
  
}  //   namespace duct

#endif
