/*  Copyright © 2016 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 17/02/16.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cmath>

/*   project headers   */
#include "bits/duct-model.h"



namespace duct {
  namespace qcd {
    
    namespace __details {
      
      
      /*   this is the proton pdf funtion at the fitting scale (HERAPDF LO)
       *   Q0 = 1GeV, alpha_s(Mz) = 0.118306 ==> as*(1+as/(2pi) K) = 0.126
       */
      static double __herapdf_func(const double *p, double x) {
        return p[0]*std::pow(x, p[1])*std::pow(1.0-x, p[2])*(1.0+p[3]*x+p[4]*x*x);
      }
      
      static void herapdf_lo_voica(double x, double *f)
      {
        constexpr double p[6][5] = {
          {2.9276,    0.6129,    4.4679,    0.0000,   12.9392},
          {1.6881,    0.6198,    3.4010,    0.0000,    0.0000},
          {0.0891,   -0.1853,    2.4870,    0.0000,    0.0000},
          {0.0891,   -0.1853,    4.8372,    0.0000,    0.0000},
          {2.5543,    0.0138,    4.4811,   -0.4657,    0.0000},
          {0.1309,   -0.1853,   16.0672,    0.0000,    0.0000}
        };
        
        f[-3] = __herapdf_func(p[5], x);
        f[-2] = __herapdf_func(p[2], x);
        f[-1] = __herapdf_func(p[3], x);
        f[ 0] = __herapdf_func(p[4], x);
        f[ 1] = __herapdf_func(p[1], x) + f[-1];
        f[ 2] = __herapdf_func(p[0], x) + f[-2];
        f[ 3] = f[-3];
        f[4]=f[-4]=f[5]=f[-5]=0.0;
      }
      
      
      static predefined herapdf_lo_init()
      {
        predefined ret;
        
        /*   unique name   */
        ret.name = "herapdflo";
        
        /*   coupling at the reference scale   */
        ret.asref = 0.118306;
        
        /*    pdf at the fitting scale   */
        ret.pdf[2212] = herapdf_lo_voica;
        
        return ret;
      }
 
      
      
      
      /*   This is the PDF function of CT14n at the fitting scale
       *   Q = 1.295GeV. The alpha_s(Mz) = 0.118.
       */
      static void ct14n_func(double x, double *f)
      {
        /*   fitted parameters   */
        static const double cumubar[8]  = {3.401536, 0.70227, 3.00938, 1.22390, 0.10933,  1.45490, 0.00000, 0.00000};
        static const double cdmdbar[8]  = {0.282512, 0.70227, 3.00938, 3.09009, 0.86623,  3.37366, 0.00000, 0.00000};
        static const double cglue[8]    = {8.712153, 0.46303, 5.25202, 1.91563,-1.09859,  0.00000, 0.00000, 0.00000};
        static const double cbar[4]     = {0.211435,-0.18450, 5.36653, 4.00000};
        static const double cdbar[6]    = {0.00000, -0.20932, 2.28705,-0.13994, 2.95568, 1.0000};
        static const double cubar[6]    = {0.00000,  0.84668, 0.62708, 1.44487, 1.26754, 0.84965};
        static const double cstrange[4] = {0.115604, -0.18450, 6.39530, 0.00000};
        
        
        /*   d - dbar & u - ubar
         *    Note that cdmdbar[i] = cumubar[i] for i = 1,2.
         */
        double y = std::sqrt(x), y2 = x, oy = 1.0-y;
        double y3 = y*y2, y4 = y2*y2, oy2 = oy*oy;
        double oy3 = oy*oy2, oy4 = oy2*oy2;
        double lx = std::log(x), lox = std::log(1.0-x);
        
        
        double aa2 = 0.1*std::log(1.0 + exp(10.0*cdmdbar[2]));
        double aa6 = 1.0 + 0.5*cdmdbar[1];
        double temp = std::exp((cdmdbar[1]-1.0)*lx + aa2*lox);
        
        double umubar = std::sinh(cumubar[3])*oy4 + std::sinh(cumubar[4])*4.0*y*oy3 + std::sinh(cumubar[5])*6.0*y2*oy2 + aa6*4.0*y3*oy + y4;
        double dmdbar = std::sinh(cdmdbar[3])*oy4 + std::sinh(cdmdbar[4])*4.0*y*oy3 + std::sinh(cdmdbar[5])*6.0*y2*oy2 + aa6*4.0*y3*oy + y4;
        
        umubar *= cumubar[0]*temp;
        dmdbar *= cdmdbar[0]*temp;
        
        
        /*    gluon    */
        double cc1 = (3.0 + 2.0*cglue[1])/3.0;
        double tempg = std::exp((cglue[1]-1.0)*lx + cglue[2]*lox);
        
        double gluon = std::sinh(cglue[3])*oy3 + std::sinh(cglue[4])*3.0*y*oy2 + cc1*3.0*y2*oy + y3;
        gluon *= tempg*cglue[0];
        
        
        /*   dbar & ubar   */
        double yb = 1.0- std::pow(oy, cbar[3]);
        double yb2 = yb*yb, oyb = 1.0-yb;
        double yb3 = yb*yb2, yb4 = yb2*yb2, oyb2 = oyb*oyb;
        double yb5 = yb*yb4, oyb3 = oyb2*oyb, oyb4 = oyb2*oyb2;
        double oyb5 = oyb4*oyb, tempbar = 0.5*std::exp((cbar[1]-1.0)*lx + cbar[2]*lox);
        
        double dbar = oyb5 + cdbar[1]*5.0*yb*oyb4 + cdbar[2]*10.0*yb2*oyb3 + cdbar[3]*10.0*yb3*oyb2 + cdbar[4]*5.0*yb4*oyb + yb5;
        double ubar = oyb5 + cubar[1]*5.0*yb*oyb4 + cubar[2]*10.0*yb2*oyb3 + cubar[3]*10.0*yb3*oyb2 + cubar[4]*5.0*yb4*oyb + cubar[5]*yb5;
        
        dbar *= tempbar*cbar[0];
        ubar *= tempbar*cbar[0];
        
        
        /*   strange   */
        double temps = (cstrange[1] - 1.0)*lx + cstrange[2]*lox;
        double strange = 0.5*cstrange[0]*std::exp(temps + cstrange[3]*(y - 0.5*x) - cstrange[1]*x);
        
        /*    returns the pdf    */
        f[-3] = x*strange;
        f[-2] = x*ubar;
        f[-1] = x*dbar;
        f[ 0] = x*gluon;
        f[ 1] = x*(dmdbar + dbar);
        f[ 2] = x*(umubar + ubar);
        f[ 3] = f[-3];
        
        f[4] = f[-4] = f[5] = f[-5] = 0.0;
      }

     
      static predefined ct14n_init()
      {
        predefined ret;
        
        /*   unique name   */
        ret.name = "ct14n";
        
        /*   coupling at the reference scale   */
        ret.asref = 0.118;
        
        /*   pdf fitting scale (transverse momentum cutoff)    */
        ret.mT = 1.295_GeV;
        
        /*    pdf at the fitting scale   */
        ret.pdf[2212] = ct14n_func;
        
        return ret;
      }
      
    }   //   namespace __details
    
    
    /*    HERA PDF
     *      It is from Voica. (This is for testing only!)
     */
    const predefined herapdflo = __details::herapdf_lo_init();

    
    /*   This is the PDF function of CT14n at the fitting scale 
     *   Q = 1.295GeV. The alpha_s(Mz) = 0.118.
     */
    const predefined ct14n = __details::ct14n_init();


    /*    List of predefined qcd inputs    */
    const predefined * default_inputs[] = {
      &ct14n,
      nullptr
    };
    
  }    //    namespace qcd
}   //   namespace duct



