/*  Copyright © 2018 Zoltán Nagy. All rights reserved.
 *  Created by Zoltan Nagy on 16.12.18.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std includes   */
#include <iostream>

/*   project headers   */
#include <bits/mc-engine-opts.h>


namespace duct {
  
  std::ostream& operator<<(std::ostream& os, const shower_engine_opts& opts)
  {
    os<<"\033[1mShower engine options:\033[0m\n";
    
    os<<"  \033[1mYou know what you are doing, don't you?\033[0m "
    <<(opts.is_insane ? "\033[31mI hope so!\033[0m" : "\033[32mThe program corrects me, what can go wrong?\033[0m")<<"\n\n"
    <<"  Transverse momentum cutoff            : "<<opts.kT_cutoff<<"GeV\n"
    <<"  Max number of the color basis vectors : "<<opts.max_number_of_color_basis_states<<"\n\n"
    <<"  Color options :\n";
    
    double tmP = 0.0;
    
    for(auto& [tmD, step_opts] : opts.color)
    {
      os<<"    On evolution range : +["<<tmP<<", "<<tmD<<"]\n";
      
      for(auto& c : step_opts)
      {
        os
        <<"      Number of steps                       : "<<c.max_number_of_steps<<"\n"
        <<"      Max color suppression                 : "<<c.max_color_suppression<<"\n"
        <<"      Type of shower evolution              : ";

        switch(c.type_of_evolution)
        {
          case evolution_type::typeI: os<<"typeI\n"; break;
          case evolution_type::typeII: os<<"typeII\n"; break;
        }
        
        os
        <<"      Order of soft corrections             : "<<c.soft_corrections.order<<"\n"
        <<"        - Real corrections                  : "<<c.soft_corrections.real<<"\n"
        <<"        - Phase corrections                 : "<<c.soft_corrections.phase<<"\n"
        <<"      Max number of the color basis vectors : "<<c.max_number_of_basis_states<<"\n"
        <<"      -----\n";
      }
      
      tmP = tmD;
    }
                                                  
    os<<"\n  Threshold options :\n"
    <<"    Threshold effects         : "<<(opts.threshold.enabled ? "Enabled\n" : "Disabled\n")
    <<"    Order of soft corrections : "<<opts.threshold.order_of_soft_corrections<<"\n";
    
    
    return os;
  }
  
  
  shower_engine_opts& shower_engine_opts::sanitize()
  {
    /*   don't run the full sanitizer if it is an insane option    */
    if(is_insane)
    {
      for(auto& copts : color)
        for(auto& c : copts.second)
          c.max_number_of_basis_states = max_number_of_color_basis_states;
      
      return *this;
    }
    
    
    /*   try to fix inconsistencies in the color options  */
    unsigned max_supp = UINT_MAX, order = UINT_MAX, real = UINT_MAX, phase = UINT_MAX;
    
    
    /*   range witout options    */
    for(auto ci = color.begin(); ci != color.end(); )
      if(ci->second.empty()) ci = color.erase(ci);
      else ++ci;
    
    
    /*   if there is no color option specified we add one and that results an LC shower   */
    if(color.empty())
      color[std::numeric_limits<double>::max()] = {
        {UINT_MAX, 0u, 0u}
      };

    
    /*   make the options consistent    */
    for(auto& copts : color)
    {
      for(auto& c : copts.second)
      {
        if(c.max_color_suppression < max_supp) max_supp = c.max_color_suppression;
        
        if(c.soft_corrections.order < order) order = c.soft_corrections.order;
        if(c.soft_corrections.real < real) real = c.soft_corrections.real;
        if(c.soft_corrections.phase < phase) phase = c.soft_corrections.phase;

        c = {c.max_number_of_steps, max_supp, {order, real, phase}, c.type_of_evolution, max_number_of_color_basis_states};
      }
      
      /*   make sure that we have full evolution on the given range    */
      copts.second.back().max_number_of_steps = UINT_MAX;
    }

    
    return *this;
  }
  
}   //   namespace duct
