/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/12/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_gauss_integral_h__
#define __math_gauss_integral_h__ 1


/*   std includes   */
#include <array>
#include <cmath>
#include <type_traits>


/*   project headers   */



namespace duct {
  
  enum gauss_integral_type : int {
    legendre = 0
  };
  
  
  template<std::size_t _Npoint>
  class gauss_integral
  {
  public:
    /*   simple sturcture to represent a basepoint with its weight    */
    struct point {
      double x, weight;
    };
    

    /*   constructors   */
    gauss_integral(gauss_integral_type type = gauss_integral_type::legendre)
    : _M_type{type}
    {
      switch(type){
        case gauss_integral_type::legendre: this->_M_gauss_legendre(); break;
        default: break;
      }
    }
    
    /*    calculate the integral   */
    template<class _Func>
    typename std::result_of<_Func(double)>::type operator()(_Func&& func) const
    {
      typename std::result_of<_Func(double)>::type tmp, retval = func(_M_base[0].x);
      
      /*    calculating the integral    */
      retval *= _M_base[0].weight;
      
      for(std::size_t k = 1; k < _Npoint; ++k) {
        tmp = func(_M_base[k].x);
        tmp *= _M_base[k].weight;
        retval += tmp;
      }
      
      return retval;
    }
    
    /*   returns the type of the integral  */
    gauss_integral_type type() const { return _M_type;}
    
    /*   return the base points and weights */
    std::array<point, _Npoint> basepoints() const { return _M_base;}
    
    
  private:
    /*   type of the integral   */
    gauss_integral_type _M_type;
    
    /*   base points and weights   */
    std::array<point, _Npoint> _M_base;
    
    
    /*   Gauss-Legenedre integral   */
    void _M_gauss_legendre();
  };
  
  
  template<std::size_t _Npoint>
  void gauss_integral<_Npoint>::_M_gauss_legendre()
  {
    constexpr double EPS = 1.0e-15;
    double z1, z, pp, p3, p2, p1;
    
    for(unsigned i = 1; i <= (_Npoint+1)/2; ++i) {
      z = std::cos(M_PI*(i-0.25)/(_Npoint+0.5));
      do {
        p1 = 1.0; p2 = 0.0;
        
        for(unsigned j = 1; j <= _Npoint; ++j) {
          p3 = p2; p2 = p1;
          p1 = ((2.0*j-1.0)*z*p2-(j-1.0)*p3)/j;
        }
        
        pp = _Npoint*(z*p1-p2)/(z*z-1.0);
        z1 = z;
        z = z1-p1/pp;
      } while(std::abs(z-z1) > EPS);
      
      _M_base[i-1].x = -z;
      _M_base[_Npoint-i].x = z;
      _M_base[i-1].weight = _M_base[_Npoint-i].weight = 2.0/((1.0-z*z)*pp*pp);
    }
  }
  
}

#endif
